@@ -1,14 +1,20 @@
 # encoding: utf-8
-require 'minitest/autorun'
-require 'stringio'
+require "minitest/autorun"
+require "stringio"
 
-class MiniSpecA < MiniTest::Spec; end
-class MiniSpecB < MiniTest::Spec; end
+class MiniSpecA < Minitest::Spec; end
+class MiniSpecB < Minitest::Test; extend Minitest::Spec::DSL; end
+class MiniSpecC < MiniSpecB; end
+class NamedExampleA < MiniSpecA; end
+class NamedExampleB < MiniSpecB; end
+class NamedExampleC < MiniSpecC; end
 class ExampleA; end
 class ExampleB < ExampleA; end
 
-describe MiniTest::Spec do
-  def assert_triggered expected = "blah", klass = MiniTest::Assertion
+describe Minitest::Spec do
+  # do not parallelize this suite... it just can"t handle it.
+
+  def assert_triggered expected = "blah", klass = Minitest::Assertion
     @assertion_count += 2
 
     e = assert_raises(klass) do
@@ -16,7 +22,7 @@ describe MiniTest::Spec do
     end
 
     msg = e.message.sub(/(---Backtrace---).*/m, '\1')
-    msg.gsub!(/\(oid=[-0-9]+\)/, '(oid=N)')
+    msg.gsub!(/\(oid=[-0-9]+\)/, "(oid=N)")
 
     assert_equal expected, msg
   end
@@ -26,10 +32,10 @@ describe MiniTest::Spec do
   end
 
   after do
-    self._assertions.must_equal @assertion_count
+    self.assertions.must_equal @assertion_count if passed? and not skipped?
   end
 
-  it "needs to be able to catch a MiniTest::Assertion exception" do
+  it "needs to be able to catch a Minitest::Assertion exception" do
     @assertion_count = 1
 
     assert_triggered "Expected 1 to not be equal to 1." do
@@ -69,25 +75,25 @@ describe MiniTest::Spec do
     @assertion_count = 2
 
     proc { raise "blah" }.must_raise RuntimeError
-    proc { raise MiniTest::Assertion }.must_raise MiniTest::Assertion
+    proc { raise Minitest::Assertion }.must_raise Minitest::Assertion
   end
 
   it "needs to catch an unexpected exception" do
     @assertion_count -= 2 # no positive
 
-    msg = <<-EOM.gsub(/^ {6}/, '').chomp
+    msg = <<-EOM.gsub(/^ {6}/, "").chomp
       [RuntimeError] exception expected, not
-      Class: <MiniTest::Assertion>
-      Message: <\"MiniTest::Assertion\">
+      Class: <Minitest::Assertion>
+      Message: <"Minitest::Assertion">
       ---Backtrace---
     EOM
 
     assert_triggered msg do
-      proc { raise MiniTest::Assertion }.must_raise RuntimeError
+      proc { raise Minitest::Assertion }.must_raise RuntimeError
     end
 
     assert_triggered "msg.\n#{msg}" do
-      proc { raise MiniTest::Assertion }.must_raise RuntimeError, "msg"
+      proc { raise Minitest::Assertion }.must_raise RuntimeError, "msg"
     end
   end
 
@@ -126,12 +132,11 @@ describe MiniTest::Spec do
                         must_output
                         must_raise
                         must_respond_to
-                        must_send
                         must_throw)
 
     bad = %w[not raise throw send output be_silent]
 
-    expected_wonts = expected_musts.map { |m| m.sub(/^must/, 'wont') }
+    expected_wonts = expected_musts.map { |m| m.sub(/^must/, "wont") }
     expected_wonts.reject! { |m| m =~ /wont_#{Regexp.union(*bad)}/ }
 
     musts.must_equal expected_musts
@@ -153,7 +158,7 @@ describe MiniTest::Spec do
   it "needs to verify binary messages" do
     42.wont_be(:<, 24).must_equal false
 
-    assert_triggered 'Expected 24 to not be < 42.' do
+    assert_triggered "Expected 24 to not be < 42." do
       24.wont_be :<, 42
     end
 
@@ -193,15 +198,16 @@ describe MiniTest::Spec do
 
     24.wont_be_close_to(42).must_equal false
 
-    assert_triggered 'Expected |42 - 42.0| (0.0) to not be < 0.001.' do
+    assert_triggered "Expected |42 - 42.0| (0.0) to not be <= 0.001." do
       (6 * 7.0).wont_be_close_to 42
     end
 
-    assert_triggered 'Expected |42 - 42.0| (0.0) to not be < 1.0e-05.' do
+    x = maglev? ? "1.0000000000000001e-05" : "1.0e-05"
+    assert_triggered "Expected |42 - 42.0| (0.0) to not be <= #{x}." do
       (6 * 7.0).wont_be_close_to 42, 0.00001
     end
 
-    assert_triggered "msg.\nExpected |42 - 42.0| (0.0) to not be < 1.0e-05." do
+    assert_triggered "msg.\nExpected |42 - 42.0| (0.0) to not be <= #{x}." do
       (6 * 7.0).wont_be_close_to 42, 0.00001, "msg"
     end
   end
@@ -211,15 +217,17 @@ describe MiniTest::Spec do
 
     24.wont_be_within_epsilon(42).must_equal false
 
-    assert_triggered 'Expected |42 - 42.0| (0.0) to not be < 0.042.' do
+    x = maglev? ? "0.042000000000000003" : "0.042"
+    assert_triggered "Expected |42 - 42.0| (0.0) to not be <= #{x}." do
       (6 * 7.0).wont_be_within_epsilon 42
     end
 
-    assert_triggered 'Expected |42 - 42.0| (0.0) to not be < 0.00042.' do
+    x = maglev? ? "0.00042000000000000002" : "0.00042"
+    assert_triggered "Expected |42 - 42.0| (0.0) to not be <= #{x}." do
       (6 * 7.0).wont_be_within_epsilon 42, 0.00001
     end
 
-    assert_triggered "msg.\nExpected |42 - 42.0| (0.0) to not be < 0.00042." do
+    assert_triggered "msg.\nExpected |42 - 42.0| (0.0) to not be <= #{x}." do
       (6 * 7.0).wont_be_within_epsilon 42, 0.00001, "msg"
     end
   end
@@ -229,15 +237,16 @@ describe MiniTest::Spec do
 
     (6.0 * 7).must_be_close_to(42.0).must_equal true
 
-    assert_triggered 'Expected |0.0 - 0.01| (0.01) to be < 0.001.' do
+    assert_triggered "Expected |0.0 - 0.01| (0.01) to be <= 0.001." do
       (1.0 / 100).must_be_close_to 0.0
     end
 
-    assert_triggered 'Expected |0.0 - 0.001| (0.001) to be < 1.0e-06.' do
+    x = maglev? ? "9.9999999999999995e-07" : "1.0e-06"
+    assert_triggered "Expected |0.0 - 0.001| (0.001) to be <= #{x}." do
       (1.0 / 1000).must_be_close_to 0.0, 0.000001
     end
 
-    assert_triggered "msg.\nExpected |0.0 - 0.001| (0.001) to be < 1.0e-06." do
+    assert_triggered "msg.\nExpected |0.0 - 0.001| (0.001) to be <= #{x}." do
       (1.0 / 1000).must_be_close_to 0.0, 0.000001, "msg"
     end
   end
@@ -247,15 +256,15 @@ describe MiniTest::Spec do
 
     (6.0 * 7).must_be_within_epsilon(42.0).must_equal true
 
-    assert_triggered 'Expected |0.0 - 0.01| (0.01) to be < 0.0.' do
+    assert_triggered "Expected |0.0 - 0.01| (0.01) to be <= 0.0." do
       (1.0 / 100).must_be_within_epsilon 0.0
     end
 
-    assert_triggered 'Expected |0.0 - 0.001| (0.001) to be < 0.0.' do
+    assert_triggered "Expected |0.0 - 0.001| (0.001) to be <= 0.0." do
       (1.0 / 1000).must_be_within_epsilon 0.0, 0.000001
     end
 
-    assert_triggered "msg.\nExpected |0.0 - 0.001| (0.001) to be < 0.0." do
+    assert_triggered "msg.\nExpected |0.0 - 0.001| (0.001) to be <= 0.0." do
       (1.0 / 1000).must_be_within_epsilon 0.0, 0.000001, "msg"
     end
   end
@@ -287,7 +296,7 @@ describe MiniTest::Spec do
   it "needs to verify instances of a class" do
     42.wont_be_instance_of(String).must_equal false
 
-    assert_triggered 'Expected 42 to not be an instance of Fixnum.' do
+    assert_triggered "Expected 42 to not be an instance of Fixnum." do
       42.wont_be_instance_of Fixnum
     end
 
@@ -299,7 +308,7 @@ describe MiniTest::Spec do
   it "needs to verify kinds of a class" do
     42.wont_be_kind_of(String).must_equal false
 
-    assert_triggered 'Expected 42 to not be a kind of Integer.' do
+    assert_triggered "Expected 42 to not be a kind of Integer." do
       42.wont_be_kind_of Integer
     end
 
@@ -352,7 +361,7 @@ describe MiniTest::Spec do
   it "needs to verify non-emptyness" do
     @assertion_count += 3 # empty is 2 assertions
 
-    ['some item'].wont_be_empty.must_equal false
+    ["some item"].wont_be_empty.must_equal false
 
     assert_triggered "Expected [] to not be empty." do
       [].wont_be_empty
@@ -390,7 +399,7 @@ describe MiniTest::Spec do
   it "needs to verify objects not responding to a message" do
     "".wont_respond_to(:woot!).must_equal false
 
-    assert_triggered 'Expected "" to not respond to to_s.' do
+    assert_triggered "Expected \"\" to not respond to to_s." do
       "".wont_respond_to :to_s
     end
 
@@ -470,7 +479,7 @@ describe MiniTest::Spec do
   end
 
   it "needs to verify using any (negative) predicate" do
-    @assertion_count -= 1 # doesn't take a message
+    @assertion_count -= 1 # doesn"t take a message
 
     "blah".wont_be(:empty?).must_equal false
 
@@ -513,7 +522,7 @@ describe MiniTest::Spec do
 
 end
 
-describe MiniTest::Spec, :let do
+describe Minitest::Spec, :let do
   i_suck_and_my_tests_are_order_dependent!
 
   def _count
@@ -542,9 +551,15 @@ describe MiniTest::Spec, :let do
 
     _count.must_equal 2
   end
+
+  it 'raises an error if the name begins with "test"' do
+    describe "let" do
+      proc { let(:test_value) { true } }.must_raise ArgumentError
+    end
+  end
 end
 
-describe MiniTest::Spec, :subject do
+describe Minitest::Spec, :subject do
   attr_reader :subject_evaluation_count
 
   subject do
@@ -560,11 +575,30 @@ describe MiniTest::Spec, :subject do
   end
 end
 
-class TestMeta < MiniTest::Unit::TestCase
-  def test_setup
-    srand 42
-    MiniTest::Unit::TestCase.reset
+class TestMetaStatic < Minitest::Test
+  def test_children
+    Minitest::Spec.children.clear # prevents parallel run
+
+    x = y = z = nil
+    x = describe "top-level thingy" do
+      y = describe "first thingy" do end
+
+      it "top-level-it" do end
+
+      z = describe "second thingy" do end
+    end
+
+    assert_equal [x], Minitest::Spec.children
+    assert_equal [y, z], x.children
+    assert_equal [], y.children
+    assert_equal [], z.children
   end
+end
+
+require "minitest/metametameta"
+
+class TestMeta < MetaMetaMetaTestCase
+  parallelize_me!
 
   def util_structure
     x = y = z = nil
@@ -596,54 +630,74 @@ class TestMeta < MiniTest::Unit::TestCase
   end
 
   def test_register_spec_type
-    original_types = MiniTest::Spec::TYPES.dup
+    original_types = Minitest::Spec::TYPES.dup
 
-    assert_equal [[//, MiniTest::Spec]], MiniTest::Spec::TYPES
+    assert_includes Minitest::Spec::TYPES, [//, Minitest::Spec]
 
-    MiniTest::Spec.register_spec_type(/woot/, TestMeta)
+    Minitest::Spec.register_spec_type(/woot/, TestMeta)
 
     p = lambda do |x| true end
-    MiniTest::Spec.register_spec_type TestMeta, &p
+    Minitest::Spec.register_spec_type TestMeta, &p
 
-    keys = MiniTest::Spec::TYPES.map(&:first)
+    keys = Minitest::Spec::TYPES.map(&:first)
 
     assert_includes keys, /woot/
     assert_includes keys, p
   ensure
-    MiniTest::Spec::TYPES.replace original_types
+    Minitest::Spec::TYPES.replace original_types
   end
 
   def test_spec_type
-    original_types = MiniTest::Spec::TYPES.dup
+    original_types = Minitest::Spec::TYPES.dup
 
-    MiniTest::Spec.register_spec_type(/A$/, MiniSpecA)
-    MiniTest::Spec.register_spec_type MiniSpecB do |desc|
+    Minitest::Spec.register_spec_type(/A$/, MiniSpecA)
+    Minitest::Spec.register_spec_type MiniSpecB do |desc|
       desc.superclass == ExampleA
     end
 
-    assert_equal MiniSpecA, MiniTest::Spec.spec_type(ExampleA)
-    assert_equal MiniSpecB, MiniTest::Spec.spec_type(ExampleB)
+    assert_equal MiniSpecA, Minitest::Spec.spec_type(ExampleA)
+    assert_equal MiniSpecB, Minitest::Spec.spec_type(ExampleB)
   ensure
-    MiniTest::Spec::TYPES.replace original_types
+    Minitest::Spec::TYPES.replace original_types
+  end
+
+  def test_name
+    spec_a = describe ExampleA do; end
+    spec_b = describe ExampleB, :random_method do; end
+
+    assert_equal "ExampleA", spec_a.name
+    assert_equal "ExampleB::random_method", spec_b.name
+  end
+
+  def test_name2
+    assert_equal "NamedExampleA", NamedExampleA.name
+    assert_equal "NamedExampleB", NamedExampleB.name
+    assert_equal "NamedExampleC", NamedExampleC.name
+
+    spec_a = describe ExampleA do; end
+    spec_b = describe ExampleB, :random_method do; end
+
+    assert_equal "ExampleA", spec_a.name
+    assert_equal "ExampleB::random_method", spec_b.name
   end
 
   def test_structure
     x, y, z, * = util_structure
 
-    assert_equal "top-level thingy", x.to_s
-    assert_equal "top-level thingy::inner thingy", y.to_s
+    assert_equal "top-level thingy",                                  x.to_s
+    assert_equal "top-level thingy::inner thingy",                    y.to_s
     assert_equal "top-level thingy::inner thingy::very inner thingy", z.to_s
 
-    assert_equal "top-level thingy", x.desc
-    assert_equal "inner thingy", y.desc
+    assert_equal "top-level thingy",  x.desc
+    assert_equal "inner thingy",      y.desc
     assert_equal "very inner thingy", z.desc
 
-    top_methods = %w(test_0001_top-level-it)
-    inner_methods1 = %w(test_0001_inner-it)
+    top_methods = %w(setup teardown test_0001_top-level-it)
+    inner_methods1 = %w(setup teardown test_0001_inner-it)
     inner_methods2 = inner_methods1 +
       %w(test_0002_anonymous test_0003_anonymous)
 
-    assert_equal top_methods,   x.instance_methods(false).sort.map(&:to_s)
+    assert_equal top_methods,    x.instance_methods(false).sort.map(&:to_s)
     assert_equal inner_methods1, y.instance_methods(false).sort.map(&:to_s)
     assert_equal inner_methods2, z.instance_methods(false).sort.map(&:to_s)
   end
@@ -651,31 +705,13 @@ class TestMeta < MiniTest::Unit::TestCase
   def test_setup_teardown_behavior
     _, _, z, before_list, after_list = util_structure
 
-    tc = z.new(nil)
+    @tu = z
 
-    tc.run_setup_hooks
-    tc.run_teardown_hooks
+    run_tu_with_fresh_reporter
 
-    assert_equal [1, 2, 3], before_list
-    assert_equal [3, 2, 1], after_list
-  end
-
-  def test_children
-    MiniTest::Spec.children.clear
-
-    x = y = z = nil
-    x = describe "top-level thingy" do
-      y = describe "first thingy" do end
-
-      it "top-level-it" do end
-
-      z = describe "second thingy" do end
-    end
-
-    assert_equal [x], MiniTest::Spec.children
-    assert_equal [y, z], x.children
-    assert_equal [], y.children
-    assert_equal [], z.children
+    size = z.runnable_methods.size
+    assert_equal [1, 2, 3] * size, before_list
+    assert_equal [3, 2, 1] * size, after_list
   end
 
   def test_describe_first_structure
@@ -689,7 +725,7 @@ class TestMeta < MiniTest::Unit::TestCase
       z = describe "second thingy" do end
     end
 
-    test_methods = ['test_0001_top level it', 'test_0002_не латинские буквы-и-спецсимволы&いった α, β, γ, δ, ε hello!!! world'].sort 
+    test_methods = ["test_0001_top level it", "test_0002_не латинские буквы-и-спецсимволы&いった α, β, γ, δ, ε hello!!! world"].sort
 
     assert_equal test_methods, [x1, x2]
     assert_equal test_methods,
@@ -700,7 +736,7 @@ class TestMeta < MiniTest::Unit::TestCase
 
   def test_structure_subclasses
     z = nil
-    x = Class.new MiniTest::Spec do
+    x = Class.new Minitest::Spec do
       def xyz; end
     end
     y = Class.new x do
@@ -712,3 +748,46 @@ class TestMeta < MiniTest::Unit::TestCase
     assert_respond_to z.new(nil), "xyz"
   end
 end
+
+class TestSpecInTestCase < MetaMetaMetaTestCase
+  def setup
+    super
+
+    @tc = self
+    @assertion_count = 1
+  end
+
+  def assert_triggered expected, klass = Minitest::Assertion
+    @assertion_count += 1
+
+    e = assert_raises klass do
+      yield
+    end
+
+    msg = e.message.sub(/(---Backtrace---).*/m, "\1")
+    msg.gsub!(/\(oid=[-0-9]+\)/, "(oid=N)")
+
+    assert_equal expected, msg
+  end
+
+  def teardown
+    msg = "expected #{@assertion_count} assertions, not #{@tc.assertions}"
+    assert_equal @assertion_count, @tc.assertions, msg
+  end
+
+  def test_expectation
+    @tc.assert_equal true, 1.must_equal(1)
+  end
+
+  def test_expectation_triggered
+    assert_triggered "Expected: 2\n  Actual: 1" do
+      1.must_equal 2
+    end
+  end
+
+  def test_expectation_with_a_message
+    assert_triggered "Expected: 2\n  Actual: 1" do
+      1.must_equal 2, ""
+    end
+  end
+end
